from typing import Any, Callable, List, TypeVar, Union

from rtry import retry

from .component import WebBrick

T = TypeVar('T', bound=Union[None, WebBrick])


def many(web_brick: T) -> List[T]:
    return web_brick.list()


class PromisedBrick:
    def __init__(self, brick, attr=None, call=None, check=None, attempts=None, timeout=None):
        self._brick = brick
        self.custom_attr = attr
        self.call = call
        self.check = check
        self.attempts = attempts or 3
        self.timeout = timeout or 1
        self.res = ''

    def __getattr__(self, item):
        if item in ['custom_attr', 'call', 'check', 'apply', 'res']:
            return object.__getattribute__(self, item)
        return PromisedBrick(self, attr=item, call=None, check=None, attempts=self.attempts, timeout=self.timeout)

    def __call__(self, *args, **kwargs):
        return PromisedBrick(self, attr=None, call=(args, kwargs), check=None, attempts=self.attempts,
                             timeout=self.timeout)

    def __eq__(self, other):
        return PromisedBrick(self, attr=None, call=None, check=('==', other), attempts=self.attempts,
                             timeout=self.timeout)

    def __ne__(self, other):
        return PromisedBrick(self, attr=None, call=None, check=('!=', other), attempts=self.attempts,
                             timeout=self.timeout)

    def __lt__(self, other):
        return PromisedBrick(self, attr=None, call=None, check=('<', other), attempts=self.attempts,
                             timeout=self.timeout)

    def __gt__(self, other):
        return PromisedBrick(self, attr=None, call=None, check=('>', other), attempts=self.attempts,
                             timeout=self.timeout)

    def __le__(self, other):
        return PromisedBrick(self, attr=None, call=None, check=('<=', other), attempts=self.attempts,
                             timeout=self.timeout)

    def __ge__(self, other):
        return PromisedBrick(self, attr=None, call=None, check=('>=', other), attempts=self.attempts,
                             timeout=self.timeout)

    def __repr__(self):
        attr_name = ''
        if self.__getattribute__('custom_attr'):
            attr = self.__getattribute__('custom_attr')
            attr_name = ('.' + attr) if attr else ''

        params = ''
        if self.__getattribute__('call'):
            call_attr = self.__getattribute__('call')
            args, kwargs = call_attr
            args = [str(item) for item in args]
            kwargs = [f'{k}={v}' for k, v in kwargs.items()]
            params = '({})'.format(', '.join(args + kwargs)) if (args + kwargs) else '()'

        check_type = ''
        if self.__getattribute__('check'):
            check_type = ' ' + self.__getattribute__('check')[0]

        check_value = ''
        if self.__getattribute__('check'):
            check_value = ' ' + self.__getattribute__('check')[1]

        res = ''
        if self.__getattribute__('res'):
            # res добавляется только при разрешении PromizedBrick и самого WebBrick
            res = ' = ' + self.__getattribute__('res')
        a = f"{self._brick}" \
            f"{attr_name}" \
            f"{params}" \
            f"{res}" \
            f"{check_type}" \
            f"{check_value}"
        return a

    def resolved_element(self):
        subj = self._brick

        if hasattr(subj, 'resolved_element'):
            if isinstance(subj.resolved_element, Callable):
                subj = subj.resolved_element()
            else:
                subj = subj.resolved_element

        if self.__getattribute__('custom_attr'):
            subj = object.__getattribute__(subj, self.custom_attr)

        if self.__getattribute__('call'):
            args, kwargs = self.__getattribute__('call')
            subj = subj(*args, **kwargs)

        self.__setattr__('res', f'{subj}')
        if self.__getattribute__('check'):
            op = self.__getattribute__('check')[0]
            other = self.__getattribute__('check')[1]
            # Что бы сохранить приоритеты для rvalue приоритетных операций над other
            if op == '==':
                return subj == other
            elif op == '!=':
                return subj != other
            elif op == '>':
                return subj > other
            elif op == '<':
                return subj < other
            elif op == '>=':
                return subj >= other
            elif op == '<=':
                return subj <= other
        return subj

    def __bool__(self):
        return retry(until=lambda x: x is True, attempts=3, timeout=1)(lambda: bool(self.resolved_element()))()

    def apply(self, swallow, attempts=2, delay=0, timeout=100):
        return retry(
            swallow=swallow, attempts=attempts, delay=delay, timeout=timeout
        )(
            lambda: self.resolved_element()
        )()


R = TypeVar('R', bound=Union[None, WebBrick, Any])


def checkable(web_brick: {R}) -> R:  # type: ignore
    return PromisedBrick(web_brick)  # type: ignore



from typing import Any, NamedTuple


class ResolverInputSet(NamedTuple):
    parent: Any = ...
    driver: Any = ...
    locator: Any = ...
    full_locator: str = ...
    strategy: str = ...
    logger: Any = ...



from dataclasses import dataclass
from datetime import datetime
from typing import Any


class LogRecord:
    def str(self):
        raise NotImplementedError

    def __str__(self):
        return self.str()

    def __repr__(self):
        return f'LogRecord("{self.str()}")'


@dataclass
class ResolutionLog(LogRecord):
    obj: Any
    component: Any

    def str(self):
        return f'{"найден" if self.component else "не найден"} {self.obj}'


@dataclass
class ResolutionErrorLog(LogRecord):
    obj: Any
    locator: Any
    error: Any

    def str(self):
        return f'Элемент не найден: {self.locator} в {self.obj}: {self.error}'


def stdout_logger(record):
    time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
    print(f'{time}: {str(record)}')



from .acesss_logger import ResolutionErrorLog, stdout_logger
from .index_locator import IndexLocator
from .resolve_result import ResolveResult
from .resolver_interface import ResolverInputSet


class NonexistingException(BaseException):
    pass


def web_resolver(waiter, ignored_exceptions=None, timeout=None, logger=stdout_logger, log_action=ResolutionErrorLog):
    if ignored_exceptions is None:
        ignored_exceptions = NonexistingException

    def resolver(resolution_input_set: ResolverInputSet):
        parent_element = resolution_input_set.parent
        locator = resolution_input_set.locator
        driver_resolve_func = resolution_input_set.strategy
        assert timeout is not None, 'Не установлен таймаут для поиска элемента на странице'

        selenium_func = {
            ResolveResult.ONE: 'find_element',
            ResolveResult.MANY: 'find_elements'
        }

        elm = None
        if isinstance(locator, IndexLocator):
            try:
                elm = parent_element[locator.index]
            except IndexError as e:
                logger(log_action(parent_element, locator, e))
            return elm

        try:
            wait = waiter(parent_element, timeout)
            elm = wait.until(lambda dr: object.__getattribute__(dr, selenium_func[driver_resolve_func])(**locator))
        except ignored_exceptions as e:
            logger(log_action(parent_element, locator, e))

        return elm

    return resolver



from .acesss_logger import ResolutionErrorLog, ResolutionLog, stdout_logger


def _value_getter(x):
    return x['value']


class WebBricksConfig:
    def __init__(self, resolver, locator_repr_extractor=None, logger=None,
                 root_locator=None, resolution_log=None, resolution_log_error=None, class_name_repr_func=None):
        self.resolver = resolver
        if root_locator is None:
            root_locator = {'by': 'css', 'value': ':root'}
        self.root_locator = root_locator

        if locator_repr_extractor is None:
            locator_repr_extractor = _value_getter
        self.locator_repr_extractor = locator_repr_extractor

        if logger is None:
            logger = stdout_logger
        self.logger = logger

        if resolution_log is None:
            resolution_log = ResolutionLog
        self.resolution_log = resolution_log

        if resolution_log_error is None:
            resolution_log_error = ResolutionErrorLog
        self.resolution_log_error = resolution_log_error

        if class_name_repr_func is None:
            def only_name_class(x):
                return x.__class__.__name__
            class_name_repr_func = only_name_class
        self.class_name_repr_func = class_name_repr_func



from .component import WebBrick
from .resolve_result import ResolveResult
from .resolver import web_resolver
from .resolver_interface import ResolverInputSet
from .utils import checkable, many
from .web_bricks_config import WebBricksConfig

__all__ = (
    'WebBrick', 'WebBricksConfig', 'web_resolver', 'ResolveResult', 'many', 'checkable', 'ResolverInputSet'
)

__version__ = "0.1.1"



import pathlib

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
INSTALL_REQUIRES = (HERE / "requirements.txt").read_text().splitlines()
TESTS_REQUIRE = (HERE / "requirements-dev.txt").read_text().splitlines()[1:]
README, DEV_README = open("README.md").read().split('# Development')
setup(
    name="web-bricks",
    version="0.1.1",
    description="Page Object constructor for UI automation",
    long_description=README,
    long_description_content_type="text/markdown",
    author="2GIS Test Labs",
    author_email="test-labs@2gis.ru",
    python_requires=">=3.8.0",
    url="https://github.com/2gis-test-labs/web-bricks",
    license="Apache-2.0",
    packages=find_packages(exclude=("tests",)),
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



from typing import List

from rtry import retry

from .index_locator import IndexLocator
from .resolve_result import ResolveResult
from .resolver_interface import ResolverInputSet
from .web_bricks_config import WebBricksConfig


class SafetyUsageError(BaseException):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class WebBrick:
    def __init__(self, parent_element, locator, driver_func=ResolveResult.ONE,
                 config: WebBricksConfig = None, resolver=None, logger=None):
        assert isinstance(parent_element, WebBrick) or config is not None, SafetyUsageError(
            f'Узел {self.__class__} прикрепляется не к дереву Component-ов, '
            f'а к драйверу parent_element={parent_element} '
            f'или забыли указать конфигурацию для корневого элемента в свойстве config={config}'
        )
        self.parent_element = parent_element
        self._locator = locator
        self._resolver = resolver
        self._driver_resolve_func_name = driver_func
        self.session_id = 'WebBrick plug session_id'
        self.config = config
        self._logger = logger

    def _parent_path(self) -> list:
        if isinstance(self.parent_element, WebBrick):
            return self.parent_element._full_path()
        return []

    def _full_path(self) -> list:
        return self._parent_path() + [self._locator]

    @property
    def locator_full_path(self):
        if not hasattr(self, '_locator_full_path_cache'):
            self._locator_full_path_cache = self._full_path()
        return self._locator_full_path_cache

    def locator_full_str_path(self):
        return self.get_root_config().locator_repr_extractor(self.locator_full_path)

    def _parent_class_path(self) -> list:
        if isinstance(self.parent_element, WebBrick):
            return self.parent_element._class_full_path()
        return []

    def _class_full_path(self) -> list:
        return self._parent_class_path() + [self.__class__.__name__]

    @property
    def class_full_path(self):
        if not hasattr(self, '_class_full_path_cache'):
            self._class_full_path_cache = self._class_full_path()
        return self._class_full_path_cache

    def __repr__(self):
        class_full_path = self.get_root_config().class_name_repr_func(self)
        locator_full_path = self.locator_full_str_path()

        many = '[]' if self._driver_resolve_func_name == ResolveResult.MANY else ''
        return f"{class_full_path}{many}('{locator_full_path}')"

    @property
    def _resolved_parent(self):
        parent_element = self.parent_element
        if isinstance(self.parent_element, WebBrick):
            parent_element = self.parent_element._resolved_current
            assert parent_element is not None, \
                f'Не найден родительский элемент {self.parent_element} для {self.__class__}:{self}'
        return parent_element

    @property
    def _resolved_current(self):
        parent_element = self._resolved_parent
        resolver = self._resolver if self._resolver else self.get_root_config().resolver
        return resolver(
            ResolverInputSet(
                parent=parent_element,
                driver=self.root_brick().driver,
                locator=self._locator,
                full_locator=self.locator_full_str_path(),
                strategy=self._driver_resolve_func_name,
                logger=self._logger
            )
        )

    @property  # type: ignore
    @retry(attempts=3, until=lambda x: x is None, swallow=AssertionError)
    def resolved_element(self):
        result = self._resolved_current
        self.log(self.get_root_config().resolution_log(self, result))
        if result is None and self._driver_resolve_func_name == ResolveResult.MANY:
            result = []
        return result

    def resolved_found_element(self, fail_msg='нет элемента на странице'):
        result = self.resolved_element
        assert (result is not None) and (result != []), f'Не нашли элемент {self}: {fail_msg}'
        return result

    def root_brick(self) -> 'WebBrick':
        parent_element = self.parent_element
        if not isinstance(parent_element, WebBrick):
            return self
        return parent_element.root_brick()

    def get_root_config(self) -> WebBricksConfig:
        root_brick = self.root_brick()
        root_brick_config = root_brick.config
        assert root_brick_config is not None, SafetyUsageError(
            f'Корневой элемент {self.__class__} должен содержать '
            f'конфиг с стратегией {root_brick.config} для поиска элементов'
        )
        # TODO cache config?
        return root_brick_config

    def log(self, record):
        self.get_root_config().logger(record)

    @property
    def driver(self):
        return self.root_brick().parent_element

    def __len__(self):
        assert self._driver_resolve_func_name == ResolveResult.MANY, SafetyUsageError(
            f'Попытка обратиться за длиной массива компонентов в {self.__class__}, '
            f'но в определении указан одиночный веб-элемент, '
            'возможно забыли добавить many? - many( WebBrick(...) )'
        )
        result = self.resolved_element
        if result is None:
            result = []
        return len(result)

    def __getitem__(self, item):
        assert self._driver_resolve_func_name == ResolveResult.MANY, SafetyUsageError(
            f'Попытка обратиться к элементу массива компонентов в {self.__class__}, '
            f'но в определении указан одиночный веб-элемент, '
            'возможно забыли добавить many? - many( WebBrick(...) )'
        )

        return self.__class__(
            parent_element=self,
            locator=IndexLocator(item),
        )

    def __iter__(self):
        assert self._driver_resolve_func_name == ResolveResult.MANY, SafetyUsageError(
            f'Попытка итерировать элементы массива компонентов в {self.__class__}, '
            f'но в определении указан одиночный веб-элемент, '
            'возможно забыли добавить many? - many( WebBrick(...) )'
        )
        return iter([self[idx] for idx in range(len(self))])

    def list(self) -> List:
        self._driver_resolve_func_name = ResolveResult.MANY
        return self  # type: ignore  # WebBrick реализует интерфейс List

    def is_equal(self, other):
        if isinstance(other, WebBrick):
            assert repr(self) != repr(other), SafetyUsageError(
                f'Сравнение описаний {self.__class__} не равносильно сравнению значений элементов. '
                f'Сравнение компонента с собой {self} всегда будет приводить к успешному результату'
            )
            return self.resolved_element is other.resolved_element
        return self.resolved_element is other

    def __eq__(self, other):
        # TODO стоит добавить __ne__, что бы корректно прокинуть not в проверки
        return self.is_equal(other)

    def __bool__(self):
        raise SafetyUsageError(
            f'Взаимодействие осуществляется с описанием элемента {self.__class__}, а не '
            f'с элементом страницы {self}, для проверки свойств используй методы и обращения к свойствам'
        )
